/* Copyright Faeem Ali 
 * email: kodgehopper@netscape.net
 *
 *----------------------------------------------------------------------
 * This file is part of quickrun 
 *
 *  quickrun is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  quickrun is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with quickrun in the file COPYING; if not, write to the Free Software
 *  Foundation, Inc.:
 *      59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *----------------------------------------------------------------------
 */

#include "complete.h"
  
/*
 * get the executables in the path
 * this is done in 2 steps:
 * 1. get directories in path from path_string
 * 2. get all executable files in those directories
 * PATH_INFO structs store the directory and executable files
 */

void get_executables()
{
   char delim[]=":";
   char *p;

#ifdef _DEBUG_
   GList *curr=NULL;
#endif
   
   /* only get new executables if the list is empty.
    * doing this only once is a little less accurate,
    * but u get better performance. 
    * given what the program is intended for (start,
    * run program and exit), this 
    * seems reasonable.
    */

   if (path_list == NULL)
     {	
	p=strtok(path_string, delim);
	while (p != NULL)
	  {
	     PATH_INFO *t;
	   
	     /* allocate memory */
	     if ((t=(PATH_INFO *)malloc(sizeof(PATH_INFO)))==NULL)
	       {
		  perror("get_path_elements(): creating PATH_INFO object");
		  exit(-1);
	       }
	     
	     /* 
	      * store the directory name and use get_path_files()
	      * to get the executable files in that directory
	      */
	     t->dir_name=strdup(p);
	     t->files=NULL;
	     get_path_files(t);
	     
	     path_list=g_list_append(path_list, t);	
	     p=strtok(NULL,delim);
	  }
	
#ifdef _DEBUG_
	curr=path_list;
	while (curr != NULL)
	  {
	     PATH_INFO *x;
	     x=(PATH_INFO *)(curr->data);
	     printf("-> path item: %s\n",x->dir_name);
	     curr=curr->next;
	  }   
#endif
     } //main if
#ifdef _DEBUG_
   else
     {
	printf("->not recalculating path variables. This is for improved speed\n");
     }
#endif
   
}//get_executables


/*
 * if filename is a regular file and is executable by this user, then return 1,
 * else return 0. if the user is root, then the necessary exceptions
 * must be made
 * match_type dictates whether we want to match regular files, directories
 * or both
 */
int is_executable(char *filename, struct stat *file_info, int match_type)
{
   /* we need the uid and gid of the current user */
   mode_t file_mode=0;
   uid_t uid=getuid();
   gid_t gid=getgid();
   
   /* check perms if the file belongs to this user and return */
   if ((file_info->st_uid==uid) || (uid==0))
     {
	file_mode=0;
	
	file_mode |= S_IXUSR;
	if (file_info->st_mode&file_mode) //if user executable bit set
	  {	     
	     switch (match_type)
	       {	     
		case MATCH_REGULAR_FILE:
		    {		  
		  if (S_ISREG(file_info->st_mode))
		    return 1;
	       }
	     break;
	     
	   case MATCH_REGULAR_AND_DIRECTORY:
	       {
		  if ((S_ISDIR(file_info->st_mode)) || (S_ISREG(file_info->st_mode)))
		    return 1;
	       }
	     break;
	  }//switch	
	  }	
     }
   
   /* if user belongs to same group and group execute perm is set */
   if ((file_info->st_gid==gid) || (gid==0))
     {
	file_mode=0;
	
	file_mode |= S_IXGRP;
	if (file_info->st_mode&file_mode)
	  {
	     switch (match_type)
	       {
		case MATCH_REGULAR_FILE:
		  if (S_ISREG(file_info->st_mode))
		    return 1;
		  break;
		  
		case MATCH_REGULAR_AND_DIRECTORY:
		  if ((S_ISDIR(file_info->st_mode)) || (S_ISREG(file_info->st_mode)))
		    return 1;		  
		  break;
	       }	          
	  }	
     }
   
   /* lastly, check other perms */
   file_mode=0;   
   file_mode |= S_IXOTH;
   if (file_info->st_mode&file_mode)
     {
	switch (match_type)
	  {
	   case MATCH_REGULAR_FILE:
	     if (S_ISREG(file_info->st_mode))
	       return 1;
	     break;
	     
	   case MATCH_REGULAR_AND_DIRECTORY:
	     if ((S_ISDIR(file_info->st_mode)) || (S_ISREG(file_info->st_mode)))
	       return 1;		  
	     break;
	  }	          
     }
      
   return 0;
}//is_executable

/*
 * gets a list of all executable files in the directory already
 * specified in t->dir_name. list of files go in t->files
 */
void get_path_files(PATH_INFO *t)
{
   DIR *dir;
   struct dirent *file_entry;
   	
   /* open directory */
   if ((dir=opendir(t->dir_name))==NULL)
     {
	/* cant access this dir...so just move on */
	return;
     }

   /* read all files in directory and add executable files to list */
   file_entry=readdir(dir);
   while (file_entry != NULL)
     {
	char *filename;
	int filename_len;
	int dir_name_len;
	struct stat file_info;
	
	/* create a full filename */
	filename_len=strlen(t->dir_name) + strlen(file_entry->d_name) + 2; //+1 for / after dirname
	if ((filename=(char *)malloc(filename_len))==NULL)
	  {
	     perror("get_path_files: allocating space for filename");
	     exit(-1);
	  }
	
	dir_name_len=strlen(t->dir_name);
	if (t->dir_name[dir_name_len-1]!='/')
	  sprintf(filename,"%s/%s",t->dir_name,file_entry->d_name);
	else
	  sprintf(filename,"%s%s",t->dir_name,file_entry->d_name);
	
	if (stat(filename,&file_info)==0)
	  {
	     /* check if executable */
	     if (is_executable(file_entry->d_name, &file_info, MATCH_REGULAR_FILE))
	       {		       
		  t->files=g_list_append(t->files,strdup(file_entry->d_name));
		  t->show_list=1;
	       }		  
	  }
	
	file_entry=readdir(dir);	     
	
	/* clean up */
	free(filename);
     }//while
   
   closedir(dir);
}//get_path_files


/*
 * frees up all memory associated with list and sets list to NULL
 * list is a list of filenames.
 */
void cleanup_list(GList **list)
{
   GList *t_list=*list;
   GList *curr=t_list;
   while (curr != NULL)
     {
	if (curr->data != NULL)
	  free(curr->data);
	curr=curr->next;
     }//while
   
   /* free up the list itself */
   g_list_free(t_list);
   *list=NULL;   
}//cleanup_list

/*
 * inserts "match" into "list" only if it isn't in the list already.
 * this is mostly intended for use with f_match, but general enough
 * to use with other lists
 */
GList *insert_match_unique(GList *list, char *match)
{
   int matched=0; //if 1, a duplicate exists
   GList *curr=list;
     
   while (curr != NULL)
     {
	char *s = (char *)(curr->data);
	if (!strcmp(s, match))
	  {
	     matched=1;
	     break;
	  }
	curr=curr->next;
     }//while   
   
   /* insert */
   if (!matched)
     list=g_list_append(list,match);
   
   return list;
}//insert_match_unique

/* 
 * match executable from the list of commands 
 * cmd is the partial executable to match from the list
 * stored in file_list
 * the matched execuables will be stored in f_match.
 * since executables can come from different directories,
 * dir_name will be set to and empty string: "". This is
 * because dir_name=NULL means the list is also empty.
 */
void match_executable(char *cmd)
{   
   int cmd_len;
   GList *curr=path_list;
   PATH_INFO *p_info=NULL;
   
   cmd_len=strlen(cmd);
   
   /* set dir_name and clean out old list if necesary */
   if (f_match.dir_name!=NULL)
     {
	free(f_match.dir_name);
	cleanup_list(&f_match.files);
     }
   f_match.dir_name=strdup("");
   
   while (curr != NULL)
     {
	GList *file_list=NULL;
	
	p_info=(PATH_INFO *)(curr->data);
	file_list=p_info->files;
	
	while (file_list != NULL)
	  {
	     char *p;
	     p=(char *)(file_list->data);
	     
	     /* if the commands match, print them out */
	     if (cmd_len==0)
	       {
		  /* match all files */
		  f_match.files=insert_match_unique(f_match.files,strdup(p));
		  f_match.show_list=1;
#ifdef _DEBUG_
		  printf("%s/%s\n",p_info->dir_name,p);
#endif		  
	       }
	     else
	       {		  
		  if (!strncmp(p,cmd,cmd_len))
		    {
		       f_match.files=insert_match_unique(f_match.files, strdup(p));
		       f_match.show_list=1;
#ifdef _DEBUG_
		       printf("%s/%s\n",p_info->dir_name,p);
#endif
		    }
	       }
	     
	     file_list=file_list->next;
	  }	
	curr=curr->next;
     }//while      
}//match_executable

/*
 * buffer is a partial filename.
 * this function finds a list of alternatives and stores it
 * in the f_match PATH_INFO struct. the data in this struct gets
 * overwritten each time it's called (if necessary)
 * this is different from finding executables. executables are
 * stored in file_list. this finds any appropriate files
 * 
 * match_type is the type of file to match. MATCH_EXECUTABLE will
 * match only executable files. the rest of the types are defined in
 * complete.h
 */
void match_file(char *buffer, int match_type)
{
   char *full_name; //duplicate of filename
   char *filename;
   char *p;
   DIR *dir;
   struct dirent *file_info;
   int filename_len; 
   
#ifdef _DEBUG_
   GList *file_list_curr;
#endif

   /* always work with a duplicate, in case we need buffer later */
   full_name=strdup(buffer);
   
   /* we need to seperate the filename and the path.
    * both may be "". if no path is specified, it defaults
    * to homedir. if no filename is specified, then all files
    * in that directory must be listed
    */
   
   /* get the path next. if the path is the same as the current path
    * in f_match, then set match_files=0, and dont get any new 
    * filenames, since we'd have them from the last time this function
    * was called. this should save a lot of time when files from the
    * same directory are being used.
    * 
    * the value of p will be used later, so it must not be changed
    */

   if ((p=strrchr(full_name,'/'))==NULL)
     {
	/* no path */
	if (f_match.dir_name!=NULL)	  
	  free(f_match.dir_name);
	
	if ((f_match.dir_name=(char *)malloc(strlen(homedir)+2))==NULL)
	  {
	     perror("match_file: error allocating memory for homedir and tailing / (1)");
	     exit(-1);
	  }	
	sprintf(f_match.dir_name,"%s/",homedir);
     }
   else
     {
	char *tmp_path=NULL;
	int path_len;
	
	path_len=(p-full_name)+1;
	
	if ((tmp_path=(char *)malloc(path_len+1))==NULL) //+1 for the NULL
	  {
	     perror("match_file: allocating memory for path");
	     exit(0);
	  }
	
	memcpy(tmp_path,full_name,path_len);
	tmp_path[path_len]='\0';

	/* overwrite dir_name in f_match if it's different and indicate that we
	 * need to get a list of files. otherwise leave things alone and save some time
	 * when we're matching filenames
	 */
	if (f_match.dir_name!=NULL)
	  {	     
	     /* if not the same */
	     free(f_match.dir_name);
	     f_match.dir_name=strdup(tmp_path);
	  }
	else
	  {
	     f_match.dir_name=strdup(tmp_path);
	  }
	free(tmp_path);		
     } //main else
   
#ifdef _DEBUG_
   printf("->match_file: using directory: %s\n",f_match.dir_name);
#endif

   /* 
    * now get a list of filenames for the directory.
    * p still points to the last "/" in full_path, so p++ points to the beginning
    * of the filename, if one has been specified.
    */
   
   /* clean up old data */
   cleanup_list(&f_match.files);	
   
   /* filename may not be a full filename, only a part of it */
   if (p!=NULL)
     {	     
	p++;
	filename=strdup(p);
	filename_len=strlen(filename);
     }
   else
     {
	/* we'll only get here if there was no '/' in buffer,
	 * which will only be the case if buffer is the 
	 * (partial) filename only
	 */
	filename=strdup(buffer);
	filename_len=strlen(filename);
     }
   
   /* now open dir and read all files in the directory but store only the names we need */
   if ((dir=opendir(f_match.dir_name))==NULL)
     {
	/* cant access directory, so reset f_match.dirname and exit */
	fprintf(stderr,"unable to access directory: \"%s\"\n",f_match.dir_name);
	free(f_match.dir_name);
	f_match.show_list=0;
	f_match.dir_name=strdup("");
	return;
     }
   
   file_info=readdir(dir);
   while (file_info != NULL)
     {
	struct stat stat_info;
	char *full_stat_name=NULL; //stat the full filename
	int full_stat_name_len=0;
	char *file=NULL;
	int len=0;
	
	full_stat_name_len=strlen(f_match.dir_name) + strlen(file_info->d_name) + 1;
	if ((full_stat_name=(char *)malloc(full_stat_name_len+1))==NULL)
	  {
	     perror("match_file: allocating memory for full name");
	     exit(-1);
	  }
	sprintf(full_stat_name,"%s/%s",f_match.dir_name,file_info->d_name);
	
	/* if the file is a directory, append a "/" to it */
	if ((stat(full_stat_name, &stat_info))==-1)
	  {
	     /* unable to stat file, so dont worry about appending "/" */
	     file=strdup(file_info->d_name);
	  }
	else
	  {	     
	     len=strlen(file_info->d_name);
	     if (S_ISDIR(stat_info.st_mode))
	       {
		  if ((file=(char *)malloc(len+2))==NULL) //+1->NULL, +1 for '/'
		    {
		       perror("match_file: allocating memory for directory file");
		       exit(-1);
		    }
		  sprintf(file,"%s/",file_info->d_name);
	       }
	     else
	       {
		  file=strdup(file_info->d_name);	     
	       }
	  }	
	
	if (filename_len==0) //no filename specified so match all files
	  {
	     /* match only the specified type of files. ignore
	      * . and ..
	      */
	     switch (match_type)
	       {	
		case MATCH_ALL:
		  if (!((!strcmp(file,"./")) || (!strcmp(file,"../"))))
		    {		       
		       f_match.files=insert_match_unique(f_match.files, strdup(file));
		       f_match.show_list=1; //we have something to show
		    }		  
		  break;
		  
		case MATCH_EXECUTABLE:
		    {
		       if (is_executable(full_stat_name, &stat_info, MATCH_REGULAR_AND_DIRECTORY))
			 {
			    if (!((!strcmp(file,"./")) || (!strcmp(file,"../"))))
			      {	
				 f_match.files=insert_match_unique(f_match.files, strdup(file));
				 f_match.show_list=1;
			      }			    
			 }		       		       
		    }
		  break;
	       }//switch
	     
	  }
	else //match selected files
	  {
	     if (!strncmp(filename,file_info->d_name,filename_len))
	       {		  
		  /* match only the specified type of files */
		  switch (match_type)
		    {	
		     case MATCH_ALL:
		       if (!((!strcmp(file,"./")) || (!strcmp(file,"../"))))
			 {
			    f_match.files=insert_match_unique(f_match.files, strdup(file));
			    f_match.show_list=1;
			 }		       
		       break;
		       
		     case MATCH_EXECUTABLE:
			 {
			    if (is_executable(full_stat_name, &stat_info, MATCH_REGULAR_AND_DIRECTORY))
			      {
				 if (!((!strcmp(file,"./")) || (!strcmp(file,"../"))))
				   {
				      f_match.files=insert_match_unique(f_match.files, strdup(file));
				      f_match.show_list=1;
				   }				 
			      }		       		       
			 }
		       break;
		    }//switch
	       }		  
	  }

	free(full_stat_name);
	free(file);
	file_info=readdir(dir);
     }//while

   closedir(dir);

   /* now it may be the case there there's no files in the file list,
    * if eg. there's no match to the partial match provided.
    * so just copy the original (partial) filename (as specified by 
    * the user) back to the list so it looks like nothing changed.
    */
   if (g_list_length(f_match.files)==0)
     {
	f_match.files=g_list_append(f_match.files,strdup(filename));
	f_match.show_list=0;
     }   
   
#ifdef _DEBUG_   

   file_list_curr=f_match.files;
   while (file_list_curr != NULL)
     {
	printf("->matched file: %s/%s\n",f_match.dir_name,(char *)(file_list_curr->data));
	file_list_curr=file_list_curr->next;
     }//while   
#endif
   
   /* clean up */
   free(full_name);
   free(filename);
}//match_file


/* 
 * this does the actual completion of the string
 * buffer is the string to complete.
 * start is where the completion should start.
 * end is offset from buffer of the cursor.
 * buffer is not modified. the new string is returned
 * 
 * basically, remove everything from start to end,
 * and replace it with f_match.dir_name and the first
 * element of f_match.files if match_type==MATCH_COMPLETE
 * or match as much of the word as possible if 
 * match_type==MATCH_PARTIAL
 */ 
char *complete_string(char *buffer, char *start, int end, int match_type)
{
   int buffer_len=0;
   int replacement_len=0;
   int new_string_len=0;
   char *replacement=NULL;
   GList *curr=f_match.files;
   
   buffer_len=strlen(buffer);
   replacement_len=strlen(f_match.dir_name) + strlen((char *)(curr->data));
   
   /* this wastes a small amount of memory, but i'm lazy to do the
    * proper calculation. the loss isnt much anyways
    */
   new_string_len=replacement_len+buffer_len;
   if ((replacement=(char *)malloc(new_string_len+1))==NULL)
     {
	perror("complete_stringerror allocating memory for replacement string");
	exit(-1);
     }
   
   memset(replacement,'\0',new_string_len+1);
   
   /* copy first part of buffer into replacement */
   if (start > buffer)
     {
	strncpy(replacement,buffer,start-buffer);	
     }
   
#ifdef _DEBUG_
   printf("->complete_string: allocated length: %d\n",new_string_len);
   printf("->complete_string: beginning: %s\n", replacement);
#endif

   /* the middle part depends on the type of matching we're performing */
   switch (match_type)
     {
      case MATCH_COMPLETE:
	  {	     	         
	     if (f_match.dir_name != NULL) //might be null from an earlier access error
	       {	
		  strcat(replacement,f_match.dir_name);
		  strcat(replacement,(char *)(curr->data));
	       }
	     else
	       {
		  free(replacement);
		  return NULL;
	       }	     
	  }
	break;
	
      case MATCH_PARTIAL:
	  {
	     if (f_match.dir_name != NULL)
	       {
		  char *p;
		  int len=0;
		  int j;
		  int copy_len=0; //number of characters to copy
		  char *s=NULL; //will point to a string in the list of options
		  GList *curr=f_match.files;
		  
		  strcat(replacement,f_match.dir_name);
		  
		  /* match as much of the words in f_match.files as 
		   * possible. first get the maximum length of all
		   * the strings in the list
		   */
		  while (curr != NULL)
		    {
		       int t_len=strlen((char *)(curr->data));
		       if (t_len > len)
			 len=t_len;
		       curr=curr->next;
		    }
		  
		  if ((p=(char *)malloc(len+1))==NULL)
		    {
		       perror("complete_string: memory for partial word match");
		       exit(-1);
		    }
		  memset(p,'\0',len+1);
		  
		  for (j=0; j<len; j++)
		    {
		       int same=-1;
		       curr=f_match.files;
		       while (curr != NULL)
			 {			    
			    s=(char *)(curr->data);
			    if (same==-1) 
			      same=s[j];
			    
			    //printf("->complete_string: s: %s, same: %c and matching: %c\n",s,same,s[j]);
			    
			    /* stop searching as soon as we find a 
			     * difference */
			    if (s[j]!=same)
			      {
				 copy_len=j;
				 j=len; //jump out of outer loop too
				 break;
			      }
			    
			    curr=curr->next;
			 }		       		       
		    }

		  /* s should point to the last string we looked at.
		   * well use this below */		  		  
		  for (j=0; j<copy_len; j++)
		    {
		       p[j]=s[j];
		    }

		  strcat(replacement,p);
		  free(p);
	       }
	     else
	       {
		  free(replacement);
		  return NULL;
	       }	     	     
	  }
	break;
     }//switch
   
#ifdef _DEBUG_
   printf("->complete_string: beg+middle: %s\n", replacement);
#endif
   
   /* add remainder of buffer */
   strcat(replacement,buffer+end);
         
#ifdef _DEBUG_
   printf("->complete_string: modified_string: \"%s\"\n",replacement);
#endif
   
   return replacement;
}//complete_string


/* this function is used when we need to complete a string, but we
 * know what the completion is. t_box is the string to complete.
 * cursor_pos is the position of the cursor (where the completion
 * will take place) and buffer is the known completion
 */
char *get_known_completion(char *buffer, int cursor_pos, char *replacement)
{
   /* lots of this code is just copied from get_completions below */
   char *t, *p;
   char *modify_start=NULL;
   char *t_buffer=NULL;
   char *t_word=NULL;
   int len=0;
   
   /* clean out what's in f_match.files and replace it with replacement.
    * leave the directory as it is since it would be properly set when
    * autocomplete was first called.
    */
   cleanup_list(&f_match.files);
   f_match.files=insert_match_unique(f_match.files, strdup(replacement));
   f_match.show_list=1;
   
   /* get portion of string to replace */
   p=t=modify_start=buffer+cursor_pos;
   t_buffer=t-1;

   while ((t_buffer > buffer) && (!isspace(*(t_buffer-1))))
     t_buffer--;			    

   modify_start=t_buffer;
   len=t-t_buffer;
   if ((t_word=(char *)malloc(len+1))==NULL)
     {
	perror("get_completions: error allocating memory for partial word (1)\n");
	exit(-1);
     }
   
   memset(t_word,'\0',len+1);
   memcpy(t_word,t_buffer,len);
   t_word[len]='\0';

   return complete_string(buffer, modify_start, cursor_pos, MATCH_COMPLETE);
}//get_known_completion

/*
 * this is the function called to get completions
 * buffer is a (possibly) incomplete command line
 * cursor_pos is the position of the cursor in buffer,
 * starting at 0.
 * command/file completion must take place given these 2 values.
 * conditions for completion:
 * 1) if the cursor is immediately after a string, match rest of string
 * 2) if cursor is alone/before a string, then match ALL strings
 * 
 * there are 2 types of strings: executables and normal files.
 * if first string/empty line, then match executables, else match
 * normal files.
 */
char *get_completions(char *buffer, int cursor_pos)
{
   char *p, *t;
   char *modify_start=NULL; /* position in buffer where modifications
			     * should start */
   int buffer_len;   
   
   buffer_len=strlen(buffer);

   /* get list of executable files from path */
   get_executables();

   p=t=buffer+cursor_pos; /* p points to the char of the cursor in the textbox
			   * textbox. 
			   */
   modify_start=p; //when matching all words, this is where we start
         
   if (buffer_len > 0) 
     {	
	/* if the previous char is a space, it means match all words */
	if (t > buffer) //can't check prev char if t == buffer
	  {	     
	     /* if the cursor is under a character, we dont match anything */
	     if ((*t=='\0') || (isspace(*t)))
	       {
		  if (isspace(*(t-1)))
		    {		       
		       /* match all words */
		       int do_match=1;
		       char *t_buffer=t-1;
		       
		       /* we need to check if we're matching a normal file or
			* an executable. work backwards. if we come to the
			* beginning of the line, match an executable.
			* if we come to another character, match a
			* normal file.
			*/
		       while (t_buffer >= buffer)
			 {
			    if (isspace(*t_buffer)) 
			      t_buffer--;
			    else
			      {
				 match_file("",MATCH_ALL);
				 do_match=0;
				 break;
			      }			    
			 }
		       if (do_match)
			 {			    
			    match_executable("");
			    if (g_list_length(f_match.files)==0)
			      {
				 match_file("",MATCH_EXECUTABLE);
			      }			    
			 }
		    }		  
		  else
		    {
		       /* match specific string */
		       char *t_buffer=t-1;
		       char *t_word=NULL;
		       int len=0;
		       int do_match=1; //a switch variable

		       /* *t is a space or null char. the previous char isnt.
			* we need to get the word to match. search backwards
			* until we come to the beginning of the word.
			* then make a copy of it and complete the filename.
			* if the beginning of the word==buffer, then search
			* for an executable. if beginning of word == space,
			* then check if it's a word or an execuable
			*/
		       while ((t_buffer > buffer) && (!isspace(*(t_buffer-1))))
			 t_buffer--;			    
		       
		       /* modifications start at where t_buffer is now,
			* so set modify_start
			*/		       
		       modify_start=t_buffer;
		       
		       len=t-t_buffer;
		       if ((t_word=(char *)malloc(len+1))==NULL)
			 {
			    perror("get_completions: error allocating memory for partial word (1)\n");
			    exit(-1);
			 }
		       
		       memset(t_word,'\0',len+1);
		       memcpy(t_word,t_buffer,len);
		       t_word[len]='\0';
		       
		       /* now check if the word to complete is an executable
			* or a normal file
			*/
		       if (t_buffer > buffer)			 
			 t_buffer--;
		       	
		       while (t_buffer > buffer)
			 {
			    if (isspace(*t_buffer)) 
			      t_buffer--;
			    else
			      {
				 match_file(t_word,MATCH_ALL);
				 do_match=0;
				 break;
			      }			    
			 }
		       if (do_match)
			 {			    
			    match_executable(t_word);
			    if (g_list_length(f_match.files)==0)
			      {
				 match_file(t_word,MATCH_EXECUTABLE);
			      }
			 }		       
		    }		  
	       }// if *t is null or space
	     else
	       {
		  /* nothing to match, just return a copy of
		   * the original string. the original is owned by the
		   * text box
		   */
		  return strdup(buffer);
	       }
	     
	  } //if (t > buffer)
	else
	  {
	     /* cursor at beginning of string so match all executables */
	     match_executable("");
	     if (g_list_length(f_match.files)==0)
	       {
		  match_file("",MATCH_EXECUTABLE);
	       }			    
	  }	
     }
   else
     {
	/* match all executables */
	match_executable("");
	if (g_list_length(f_match.files)==0)
	  {
	     match_file("",MATCH_EXECUTABLE);
	  }			    	
     }

   /* at this point, f_match.files contains the list of files
    * we're interested in. if there's exactly 1 file in the list
    * we modify buffer and return it. if there's more than one file,
    * we cant modify buffer so just return "". returning NULL means error.
    * the else portion might be NULL.
    */
   if (g_list_length(f_match.files) > 1)
     {	
	/* match as much of the string as possible.
	 * see the big note below for some info
	 * on why we're not free()'ing here
	 */
	buffer=complete_string(buffer, modify_start, cursor_pos,
			       MATCH_PARTIAL);
	return buffer;
     }
   else
     {
	if (g_list_length(f_match.files)==0)
	  {
	     return NULL; //some error occurred. mebe a permission denied etc.
	  }		       
	else
	  {
	     /* char *t=buffer; */
	     buffer=complete_string(buffer,modify_start,cursor_pos, 
				    MATCH_COMPLETE);
	     
	     /***********NOTE TO OTHER PROGRAMMERS*************/
	     /* hmmm...ok. we can't free t (which was buffer) because gtk
	      * is still in control of it.
	      * however, when we reset the text in the text box,
	      * memory management for t will be handled by gtk.
	      * so not calling free() here shouldnt be a problem. If this
	      * function is used in some other code, make sure free(t)
	      * is called if necessary or else there'll be a memory leak.
	      * 
	      * also, once buffer has been used, it needs to be free()'ed
	      * as well. it is free()'ed in this code, just in the calling
	      * function, not here.
	      */
	     
	     /* free(t) */
	     return buffer;
	  } //inner else
     } //main else
}

