/* Copyright Faeem Ali 
 * email: kodgehopper@netscape.net
 *
 *----------------------------------------------------------------------
 * This file is part of quickrun 
 *
 *  quickrun is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  quickrun is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with quickrun in the file COPYING; if not, write to the Free Software
 *  Foundation, Inc.:
 *      59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *----------------------------------------------------------------------
 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <glib.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>

typedef struct _PATH_INFO
{
   char *dir_name;
   GList *files;
} PATH_INFO;

char *path_string;
char *home_dir;
PATH_INFO f_match; //statically stores filename match information
GList *path_list=NULL; // list of elements in the path


/* prototypes */
void get_path_files(PATH_INFO *t);
  
/*
 * get the executables in the path
 * this is done in 2 steps:
 * 1. get directories in path from path_string
 * 2. get all executable files in those directories
 * PATH_INFO structs store the directory and executable files
 */

void get_executables()
{
   char delim[]=":";
   char *p;

#ifdef _DEBUG_
   GList *curr=NULL;
#endif
   
   p=strtok(path_string, delim);
   while (p != NULL)
     {
	PATH_INFO *t;
	
	/* allocate memory */
	if ((t=(PATH_INFO *)malloc(sizeof(PATH_INFO)))==NULL)
	  {
	     perror("get_path_elements(): creating PATH_INFO object");
	     exit(-1);
	  }

	/* 
	 * store the directory name and use get_path_files()
	 * to get the executable files in that directory
	 */
	t->dir_name=strdup(p);
	t->files=NULL;
	get_path_files(t);
	
	path_list=g_list_append(path_list, t);	
	p=strtok(NULL,delim);
     }
   
#ifdef _DEBUG_
   curr=path_list;
   while (curr != NULL)
     {
	PATH_INFO *x;
	x=(PATH_INFO *)(curr->data);
	printf("-> path item: %s\n",x->dir_name);
	curr=curr->next;
     }   
#endif
   
}//get_path_elements


/*
 * if filename is a regular file and is executable by this user, then return 1,
 * else return 0
 */
int is_executable(char *filename, struct stat *file_info)
{
   /* we need the uid and gid of the current user */
   mode_t file_mode=0;
   uid_t uid=getuid();
   gid_t gid=getgid();
   
   /* check perms if the file belongs to this user and return */
   if (file_info->st_uid==uid)
     {
	file_mode=0;
	
	file_mode |= S_IXUSR;
	if ((S_ISREG(file_info->st_mode)) && (file_info->st_mode&file_mode))
	  return 1;
     }
   
   /* if user belongs to same group and group execute perm is set */
   if (file_info->st_gid==gid)
     {
	file_mode=0;
	
	file_mode |= S_IXGRP;
	if ((S_ISREG(file_info->st_mode)) && (file_info->st_mode&file_mode))
	  return 1;
     }
   
   /* lastly, check other perms */
   file_mode=0;   
   file_mode |= S_IXOTH;
   if ((S_ISREG(file_info->st_mode)) && (file_info->st_mode&file_mode))
     return 1;
   
   return 0;
}//is_executable

/*
 * gets a list of all executable files in the directory already
 * specified in t->dir_name. list of files go in t->files
 */
void get_path_files(PATH_INFO *t)
{
   DIR *dir;
   struct dirent *file_entry;
   	
   /* open directory */
   if ((dir=opendir(t->dir_name))==NULL)
     {
	/* cant access this dir...so just move on */
	return;
     }

   /* read all files in directory and add executable files to list */
   file_entry=readdir(dir);
   while (file_entry != NULL)
     {
	char *filename;
	int filename_len;
	int dir_name_len;
	struct stat file_info;
	
	/* create a full filename */
	filename_len=strlen(t->dir_name) + strlen(file_entry->d_name) + 2; //+1 for / after dirname
	if ((filename=(char *)malloc(filename_len))==NULL)
	  {
	     perror("get_path_files: allocating space for filename");
	     exit(-1);
	  }
	
	dir_name_len=strlen(t->dir_name);
	if (t->dir_name[dir_name_len-1]!='/')
	  sprintf(filename,"%s/%s",t->dir_name,file_entry->d_name);
	else
	  sprintf(filename,"%s%s",t->dir_name,file_entry->d_name);
	
	if (stat(filename,&file_info)==0)
	  {
	     /* check if executable */
	     if (is_executable(file_entry->d_name, &file_info))
	       {		       
		  t->files=g_list_append(t->files,strdup(file_entry->d_name));
	       }		  
	  }
	
	file_entry=readdir(dir);	     
	
	/* clean up */
	free(filename);
     }//while
   
   closedir(dir);
}//get_path_files


/* match executable from the list of commands */
void match_executable(char *cmd)
{   
   int cmd_len;
   GList *curr=path_list;
   PATH_INFO *p_info=NULL;
   
   cmd_len=strlen(cmd);
   
   while (curr != NULL)
     {
	GList *file_list=NULL;
	
	p_info=(PATH_INFO *)(curr->data);
	file_list=p_info->files;
	
	while (file_list != NULL)
	  {
	     char *p;
	     p=(char *)(file_list->data);
	     
	     /* if the commands match, print them out */
	     if (!strncmp(p,cmd,cmd_len))
	       {
		  printf("%s/%s\n",p_info->dir_name,p);
	       }
	     
	     file_list=file_list->next;
	  }	
	curr=curr->next;
     }//while      
}//match_executable


/*
 * frees up all memory associated with list and sets list to NULL
 * list is a list of filenames.
 */
void cleanup_list(GList **list)
{
   GList *t_list=*list;
   GList *curr=t_list;
   while (curr != NULL)
     {
	if (curr->data != NULL)
	  free(curr->data);
	curr=curr->next;
     }//while
   
   /* free up the list itself */
   g_list_free(t_list);
   *list=NULL;   
}//cleanup_list

/*
 * buffer is a partial filename.
 * this function finds a list of alternatives and stores it
 * in the f_match PATH_INFO struct. the data in this struct gets
 * overwritten each time it's called (if necessary)
 */
void match_file(char *buffer)
{
   char *full_name; //duplicate of filename
   char *filename;
   char *p;
   int match_files=1; /* if 0, we already have a valid list of files eg. 
		       * when we're processing the same directory twice
		       */
   
#ifdef _DEBUG_
   GList *file_list_curr;
#endif

   /* always work with a duplicate, in case we need buffer later */
   full_name=strdup(buffer);
   
   /* we need to seperate the filename and the path.
    * both may be "". if no path is specified, it defaults
    * to home_dir. if no filename is specified, then all files
    * in that directory must be listed
    */
   
   /* get the path next. if the path is the same as the current path
    * in f_match, then set match_files=0, and dont get any new 
    * filenames, since we'd have them from the last time this function
    * was called. this should save a lot of time when files from the
    * same directory are being used
    * 
    * the value of p will be used later, so it must not be changed
    */
   if ((p=strrchr(full_name,'/'))==NULL)
     {
	/* no path */
	if (f_match.dir_name!=NULL)
	  {	     
	     if (!strcmp(f_match.dir_name, home_dir))
	       {
		  match_files=0;
	       }
	     else
	       {		  
		  free(f_match.dir_name);
		  f_match.dir_name=strdup(home_dir);
		  match_files=1;
	       }	     
	  }
	else
	  {
	     f_match.dir_name=strdup(home_dir);
	     match_files=1;
	  }	
     }
   else
     {
	char *tmp_path=NULL;
	int path_len;
	
	path_len=(p-full_name)+1;
	
	if ((tmp_path=(char *)malloc(path_len+1))==NULL) //+1 for the NULL
	  {
	     perror("match_file: allocating memory for path");
	     exit(0);
	  }
	
	memcpy(tmp_path,full_name,path_len);
	tmp_path[path_len-1]='\0';
	
	/* overwrite dir_name in f_match if it's different and indicate that we
	 * need to get a list of files. otherwise leave things alone and save some time
	 * when we're matching filenames
	 */
	if (f_match.dir_name!=NULL)
	  {
	     if (!strcmp(tmp_path,f_match.dir_name))
	       {
		  match_files=0;
	       }
	     
	     /* if not the same */
	     free(f_match.dir_name);
	     f_match.dir_name=strdup(tmp_path);
	     match_files=1;
	  }
	else
	  {
	     f_match.dir_name=strdup(tmp_path);
	     match_files=1;
	  }
	free(tmp_path);		
     } //main else
   

   /* 
    * now get a list of filenames for the directory in f_match if match_files==1 
    * p still points to the last "/" in full_path, so p++ points to the beginning 
    * of the filename, if one has been specified.
    */
   if (match_files==1)
     {
	DIR *dir;
	struct dirent *file_info;
	int filename_len; 
	
	/* clean up old data */
	cleanup_list(&f_match.files);
	
	/* filename may not be a full filename, only a part of it */
	p++;
	filename=strdup(p);
	filename_len=strlen(filename);
	
	/* now open dir and read all files in the directory but store only the names we need */
	if ((dir=opendir(f_match.dir_name))==NULL)
	  {
	     /* cant access directory, so reset f_match.dirname and exit */
	     free(f_match.dir_name);
	     f_match.dir_name=NULL;
	     return;
	  }
	
	file_info=readdir(dir);
	while (file_info != NULL)
	  {
	     if (filename_len==0) //no filename specified so match all files
	       {
		  f_match.files=g_list_append(f_match.files,strdup(file_info->d_name));		  
	       }
	     else //match selected files
	       {
		  if (!strncmp(filename,file_info->d_name,filename_len))
		    {		       
		       f_match.files=g_list_append(f_match.files,strdup(file_info->d_name));
		    }		  
	       }
	     	     
	     file_info=readdir(dir);
	  }//while
	
	closedir(dir);
     }
   
#ifdef _DEBUG_
   file_list_curr=f_match.files;
   while (file_list_curr != NULL)
     {
	printf("->matched file: %s/%s\n",f_match.dir_name,(char *)(file_list_curr->data));
	file_list_curr=file_list_curr->next;
     }//while   
#endif
   
   /* clean up */
   free(full_name);
}//match_file

int main(int argc, char *argv[])
{
   if (argc < 2)
     {
	printf("Usage: complete <beginning_of_cmd> <beginning_of_filename>\n");
	return 1;
     }
   
   
   /* get path variable */
   if ((path_string=getenv("PATH"))==NULL)
     {
	fprintf(stderr,"error determining path from $PATH\n");
	exit(-1);       
     }
   
#ifdef _DEBUG_   
   printf("->path: %s\n",path_string);
#endif
   
   /* get home dir */
   if ((home_dir=getenv("PATH"))==NULL)
     {
	fprintf(stderr,"error getting home directory. setting to \"/\"\n");
	home_dir=strdup("/");
     }

#ifdef _DEBUG_
   printf("->home directory: %s\n",home_dir);
#endif

   
   /* get list of executable files from path */
   get_executables();
         
   /* match executable */
   match_executable(argv[1]);
   
   /* match filename */
   match_file(argv[2]);
   
   return 0;
}
